from utils.mylog import logger, log_levels
from lbm import LBM
import click
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.animation as animation
import matplotlib.colors
import logging
from lbmweights import Lattice


pause = True
pause_immediately = False


def onClick(event):
    global pause_immediately
    global pause
    if event.key == " ":
        pause ^= True

    if event.key == "n":
        pause = False
        pause_immediately = True


class CloseEvent:
    def __init__(self, lbm):
        self.lbm = lbm

    def __call__(self, event):
        """TODO: This is a dirty Hack!
        This function get's called when the user presses the x button on the canvas. It can be used for
        post processing"""
        logger.info("---- Execution of LBM stopped. Output Metadata")
        from utils.utils import Timer

        # overkill because I only save one value !
        for (func_name, time) in [x for x in sorted(Timer.times.items(), key=lambda x: x[1])]:
            logger.warning("{} execution time: {}".format(func_name, time))


@click.command()
@click.option("--init", type=str, default="taylor-green")
@click.option("--debug", type=str, default="warning")
@click.option("--animate", type=bool, default=True)
@click.option("--skip", type=int, default=0)
@click.option("--lattice", type=str, default="D2Q9")
@click.option("--pystencils", type=bool, default=False)  # TODO == True eventually
@click.option("--size", type=(int, int), default=(100, 100))
@click.option("--test", type=bool, default=False)
def main(init, debug, animate, skip, lattice, pystencils, size, test):
    """
    :param init: Fill type for the initial condition
    :param debug: Sets the logger to the corresponding level. Possible values are "warning", "info" or "debug".
    :param lattice: Type of Lattice. Naming Convention is upppercase and the established names from literature
        are taken. For a full list of the lattices look at lbmweights.Lattice.BY_NAMES. Only 2d is implemented.

    :param animate: Show animation
    :param skip: Int. Skip the plotting of the first <skip> iteration. This means that the colorbar evaluated only
        after those iterations and the colors are more sensitive.
        #TODO this is lazy and can be done without skipping the plots of the first few iterations

    :param test: Used for debugging resons. Implement two versions of LBM and compare them
    :return:
    """


    lattice = Lattice.from_name(lattice)
    if lattice._dimension != 2:
        raise ValueError("This Script is not suited for non two-dimensional simulations")

    logger.setLevel(log_levels[debug.lower()])
    lbm = LBM(
        size,
        lattice=lattice,
        omega=1/0.5135043208296701,
        plot_type="density",
        fill_mode=init,
        pystencils=pystencils,
        test=test,
    )
    lbm.update_density()
    lbm.update_velocity()

    # evaluate the first <skip> time step iterations
    for i in range(skip):
        lbm.iterate()

    # Postprocessing
    on_close = CloseEvent(lbm)

    # plotting
    fig = plt.figure()
    fig.canvas.mpl_connect("close_event", on_close)
    gs = fig.add_gridspec(2, 2)
    ax1 = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[1, 0])
    ax3 = fig.add_subplot(gs[0, 1])
    ax4 = fig.add_subplot(gs[1, 1])

    # First graphic, standard plot_type --> density
    ax1.set_title(lbm.plot_type)
    mat1 = ax1.matshow(lbm.plot_data(), cmap=cm.GnBu)
    ax1.set_yticklabels([])
    ax1.set_xticklabels([])
    fig.colorbar(mat1, ax=ax1)

    # Second Graphic, Direction of velocity
    # TODO Bug argmax([0,..,0]) = 0 should be 4
    cmap = matplotlib.colors.ListedColormap(
        [
            "#33cccc",
            "#0099ff",
            "#3333cc",
            "#33cc33",
            "#ffffff",
            "#ff00ff",
            "#cccc00",
            "#ff3300",
            "#b30006",
        ]
    )
    boundaries = [-0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5]
    norm = matplotlib.colors.BoundaryNorm(boundaries, cmap.N, clip=True)
    ax2.set_title("Velocity Direction")
    ax2.set_yticklabels([])
    ax2.set_xticklabels([])
    mat2 = ax2.matshow(
        lbm.plot_data("velocity direction"),
        label="Velocity Direction",
        cmap=cmap,
        norm=norm,
    )

    # Third graphic, total velocity
    ax3.set_title("relative velocity")
    mat3 = ax3.matshow(lbm.plot_data("total velocity"), cmap=cm.cool)
    ax3.set_yticklabels([])
    ax3.set_xticklabels([])
    fig.colorbar(mat3, ax=ax3)

    ax4.set_title("Analytical Solution")
    mat4 = ax4.matshow(lbm.plot_data("Analytical Solution"), cmap=cm.cool)
    ax4.set_yticklabels([])
    ax4.set_xticklabels([])
    fig.colorbar(mat4, ax=ax4)
    # Quiver Plot, takes long to calculate
    # Fourth graphic, direction of velocity using quiverplot
    # ax4.set_title("Normalized Quiverplot")
    # ax4.set_aspect("equal")
    # X, Y, U, V = lbm.plot_data("quiver")
    # mat4 = ax4.quiver(X, Y, U, V, units="xy", scale=1)
    # fig.colorbar(mat4, ax=ax4)

    # Start of animation
    def iteration(frame):
        global pause
        global pause_immediately
        if pause:
            return
        # TODO this should be ==0 but it crashes
        if frame == -1:
            logger.debug("im getting called in the beginning")
            lbm.reset()
        logger.debug("entering iteration func in main Frame: {}".format(frame))
        lbm.iterate()
        mat1.set_data(lbm.plot_data())
        mat2.set_data(lbm.plot_data("velocity direction"))
        mat3.set_data(lbm.plot_data("total velocity"))
        mat4.set_data(lbm.plot_data("Analytical Solution"))

        # Skip one frame by pressing n
        if pause_immediately:
            pause_immediately = False
            pause ^= True

    if animate:
        fig.canvas.mpl_connect("key_press_event", onClick)
        ani = animation.FuncAnimation(
            fig, iteration, frames=100, interval=20, repeat=True
        )

    plt.show()


if __name__ == "__main__":
    main()
