from lbmweights import Lattice
import numpy as np
from utils.mylog import logger
from utils.utils import Timer, l2norm, taylor_green, add_ghost_layer
timer = Timer.timer



"""
The Lattice Boltzmann Method - Principles and Practice Book.
"""


class LBM:
    def __init__(
        self,
        size,
        lattice,
        pystencils,
        omega=1,
        plot_type="total_velocity",
        fill_mode="zero",
        hurry=False,
        test=False,
    ):
        """
        :param size: dimensions of the simulation in x and y direction
        :param lattice: Lattice object from lbmweights.
        :param omega: Inverse of relaxation time (for delta_t =1)
        :param plot_type: Declares what the default plot type of lbm will be. For a list of all plot types look at
            `func:LBM.plot_data()`
        :param fill_mode: Different versions of initializing the density distribution (data). For a list of all
            implemented fill modes look at `:func:LBM.fill`
        :param hurry: Set to true to sacrifice computation time for output about accuracy and convergence
        :param test: Used for debugging. Test if changes made to the code result in changes
        """
        self.delta_t = 1
        self.omega = omega
        self.relaxation_time = self.delta_t / omega  # this may be deprecated

        assert self.omega == self.delta_t / self.relaxation_time  # this as well
        self.iteration = 0
        self.accuracy = 0
        self.pystencils = pystencils
        self.test = test
        self.lattice = lattice

        # weights and velocities from lattice
        c_s_sq, weights, velocity = lattice.velocity_set()
        self.weights = np.array(weights, dtype=np.float64)
        self.c_s_sq = float(c_s_sq)
        self.c_ix = velocity[0]
        self.c_iy = velocity[1]

        # !!!!!
        #self.relaxation_time = 0.1 / self.c_s_sq + 0.5
        self.omega = 1 / self.relaxation_time

        self.plot_type = plot_type
        self.shape = size + (len(self.weights),)
        logger.debug("Initialized lbm with shape: {}".format(self.shape))
        self._data = np.zeros(self.shape, dtype=np.float64)
        self._f_eq = np.zeros_like(self._data)
        self._density = np.zeros(size)
        self._initial_density = 0
        self._fill_mode = fill_mode
        self._hurry = hurry
        self.vel_x = np.zeros(size)
        self.vel_y = np.zeros(size)

        if self.pystencils:
            import pystencils_kernels as pystencils_kernels

            logger.debug("Pystencils activated")
            self._update_density_kernel = pystencils_kernels.update_density_kernel(
                self.q
            )
            self._update_velocity_kernel = pystencils_kernels.update_velocity_kernel(
                self.q, self.c_ix, self.c_iy
            )
            self._update_f_equilibrium_kernel = pystencils_kernels.update_f_equilibrium_kernel(
                Q=self.q, weights=self.weights, c_ix=self.c_ix, c_iy=self.c_iy, c_s_sq=self.c_s_sq, test=self.test
            )
            self._collision_kernel = pystencils_kernels.collision_kernel(
                self.q, self.omega
            )
            self._propagation_kernel = pystencils_kernels.propagation_kernel(
                size=size, Q=self.q, c_ix=self.c_ix, c_iy=self.c_iy
            )
        else:
            if len(self.weights) not in [9, 17]:
                raise ValueError("Only D2Q9 and D2V17 are supported for NumPy Implementations if NumPy is used")

        self.fill()
        self._initial_density = self.density.sum()
        self._initial_ddf = np.copy(self.data)

        logger.debug("exiting init function")

    @property
    def initial_density(self):
        return self._initial_density

    @property
    def q(self):
        return self.f_eq.shape[2]

    @property
    def fill_mode(self):
        return self._fill_mode

    @property
    def f_eq(self):
        return self._f_eq

    @f_eq.setter
    def f_eq(self, f_eq_new):
        logger.info("New value set for f_eq")
        self._f_eq = f_eq_new

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_data):
        self._data = new_data

    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, new_density):
        self._density = new_density

    @property
    def velocity(self):
        return self.vel_x, self.vel_y

    def reset(self):
        logger.debug("Resetting Field")
        self.data = self.initial_density
        self.iteration = 0

    def plot_data(self, plot_type=None):
        """
        :return: Data that gets plotted by matplot
        """
        if not plot_type:
            return self.plot_data(self.plot_type)
        if plot_type == "total velocity":
            logger.debug(
                "Plotting type total velocity outputs 2d data array tath indicates the total velocity"
            )
            vel_x, vel_y = self.velocity
            total = vel_x ** 2 + vel_y ** 2
            return total
        elif plot_type == "velocity direction":
            # Description what
            vel = self.data[:, :] - self.weights
            # a 0.1 discrepancy from weights at the edges (NE, NW etc) has more influence than 0.1 in the middle
            weighting = [1 - w for w in self.weights]
            vel[:, :] *= weighting
            return vel.argmax(axis=2)
        elif plot_type == "density":
            return self.density
        elif plot_type == "quiver":
            vel = self.data[:, :] - self.weights


            vel[abs(vel) < 0.0001] = 0

            v_0 = self.data[:, :, 0] - self.weights[0]

            X, Y = np.meshgrid(np.arange(0, self.shape[0]), np.arange(0, self.shape[1]))

            # which direction does differ the most from equilibrium
            am = vel.argmax(axis=2)
            am[np.logical_and(am == 0, v_0 == 0)] = 4
            U = self.c_ix[am] * 0.8
            V = self.c_iy[am] * 0.8
            return X, Y[::-1, :], U, V

        elif plot_type == "Analytical Solution":
            density, vel_x, vel_y = taylor_green(self.shape[:2], self.iteration, self.relaxation_time, self.c_s_sq, self.test)
            vel_abs = vel_x ** 2 + vel_y ** 2
            return vel_abs

    def get_density_distribution(self, x, y):
        """Return density distrubution for one specific value"""
        return self.data[x, y].reshape((3, 3))

    def check_initial_density(self):
        discrepancy = np.abs(self.density.sum() - self.initial_density)/self.initial_density
        if discrepancy > 0.001 and self.iteration > 0:
            logger.warning("Density is not conserved anymore. Distance: {}".format(discrepancy))

    def update_density(self):
        logger.debug("density is being updated")
        if self.pystencils:
            self._update_density_kernel(data=self.data, density=self.density)
        else:
            self.density = self.data.sum(axis=2)

        if self.density.min() <= 0:
            logger.warning("Negative or zero value in density")

        self.check_initial_density()
        return self.density

    def fill(self):
        logger.info("Filling data unsing the fill mode {}".format(self.fill_mode))
        fill_value = 0.05
        self.data[:, :] = self.weights
        mx = int(len(self.data) / 2)
        my = int(len(self.data[mx]) / 2)

        if self.fill_mode == "double river":
            self.density[:, :] = 1
            self.vel_x[:, :] = fill_value
            self.vel_x[len(self.data) // 3 : len(self.data) // 3 * 2] = -1 * fill_value
            self.vel_y = (
                fill_value
                * 0.1
                * np.random.rand(self.data.shape[0], self.data.shape[1])
            )

            self.update_f_equilibrium()
            self.data = self.f_eq

        elif self.fill_mode == "drop":
            if self.q != 9:
                logger.warning("Not implemented for q != 9 yet")
            self.data[mx - 3 : mx + 3, my - 3 : my + 3, :] += (
                np.array([0, 0, 0, 0, 9, 0, 0, 0, 0]) * fill_value
            )

            self.update_density()
            self.update_velocity()
            self.update_f_equilibrium()
            self.data = self.f_eq

            self.data[mx - 3 : mx + 3, my - 3 : my + 3, :] += (
                np.array([0, 0, 0, 0, 9, 0, 0, 0, 0]) * fill_value
            )
        elif self.fill_mode == "taylor-green":
            # p. 660

            (density, vel_x, vel_y) = taylor_green(self.data.shape, 0, self.relaxation_time, self.c_s_sq, self.test)
            self.density = density
            self.vel_x = vel_x
            self.vel_y = vel_y
            self.update_f_equilibrium()
            self.data = self.f_eq

        elif self.fill_mode == "equilibrium":
            self.data = np.zeros_like(self.data)
            return

        elif self.fill_mode == "double source":
            self.data[mx - 10 : mx, my + 20 : my + 25] += (
                np.array([0, 0, 0, 0, 0, 0, 0, 0, 2]) * fill_value
            )
            self.data[mx - 5 : mx + 5, my + 30 : my + 35] += (
                np.array([1, 0, 0, 0, 0, 0, 0, 0, 0]) * fill_value
            )

        elif self.fill_mode == "mei river":
            # Initializing using mei  et al.'s Algorithm p. 224
            # 1. Initialise Population using 5.84
            self.density[:, :] = 1
            self.vel_x[:mx, :] = -1
            self.vel_x[mx:, :] = 1
            self.vel_y = np.zeros_like(self.vel_x)
            # self.vel_x[mx:mx+2, my:my+2] = [[1.8, 1.8], [-1.8, -1.8]]
            # self.vel_y[mx:mx+2, my:my+2] = [[1.8, -1.8], [1.8, -1.8]]
            self.update_f_equilibrium()
            self.data = self.f_eq
            init = np.zeros_like(self.data)
            counter = 0
            while l2norm(init, self.data) > 10e-07:
                counter += 1
                logger.debug("Mei Initialisation scheme iteration {}".format(counter))
                init = np.copy(self.data)

                # 2. Compute the local density
                self.update_density()

                # 3. Perform Collision with velocity unchanged
                self.update_f_equilibrium()
                self.do_collision(omega=1)

                # 4. Propagate
                self.do_propagation()
            logger.debug("Maximum value in density: {}".format(self.density.max()))

    def update_velocity(self):
        """
        Updates the velocity of the data by calculating the momentum density and dividing by density
        Afterwards they can be accessed using the velocity property
        """

        def calc_vel_x(self):
            if self.q == 9:
                return ((self.data[:, :, 8] + self.data[:, :, 3] + self.data[:, :, 6])
                        - (self.data[:, :, 7] + self.data[:, :, 2] + self.data[:, :, 5])
                       ) / self.density
            if self.q == 17:
                return (
                               (
                                       self.data[:, :, 3]
                                       + self.data[:, :, 8]
                                       + self.data[:, :, 6]
                                       + 2 * self.data[:, :, 12]
                                       + 2 * self.data[:, :, 10]
                                       + 3 * self.data[:, :, 15]
                               )
                               - (
                                       self.data[:, :, 2]
                                       + self.data[:, :, 7]
                                       + self.data[:, :, 5]
                                       + 2 * self.data[:, :, 11]
                                       + 2 * self.data[:, :, 9]
                                       + 3 * self.data[:, :, 14]
                               )
                       ) / self.density

        def calc_vel_y(self):
            if self.q == 9:
                return (
                            (self.data[:, :, 7] + self.data[:, :, 4] + self.data[:, :, 8])
                               - (self.data[:, :, 5] + self.data[:, :, 1] + self.data[:, :, 6])
                       ) / self.density
            elif self.q == 17:
                return (
                           (
                                   self.data[:, :, 4]
                                   + self.data[:, :, 7]
                                   + self.data[:, :, 8]
                                   + 2 * self.data[:, :, 11]
                                   + 2 * self.data[:, :, 12]
                                   + 3 * self.data[:, :, 16]
                           )
                           - (
                                   self.data[:, :, 1]
                                   + self.data[:, :, 5]
                                   + self.data[:, :, 6]
                                   + 2 * self.data[:, :, 9]
                                   + 2 * self.data[:, :, 10]
                                   + 3 * self.data[:, :, 13]
                           )
                   ) / self.density

        if self.pystencils:
            vel = np.zeros(
                (self.vel_x.shape[0], self.vel_x.shape[1], 2), dtype=np.float64
            )
            self._update_velocity_kernel(
                data=self.data, density=self.density, velocity=vel
            )
            self.vel_x = vel[:, :, 0]
            self.vel_y = vel[:, :, 1]
        else:
            self.vel_x = calc_vel_x(self)
            self.vel_y = calc_vel_y(self)

    def update_macroscopic_fields(self):
        """
        Gets called in every iteration to update the value of density and velocity according to the new
        Density Distribution Function.

        :return:
        """
        self.update_density()
        self.update_velocity()

    def update_f_equilibrium(self):
        """
        Update equilibrium density funcition.
        Only works for D2Q9
        """
        if self.pystencils:
            self._update_f_equilibrium_kernel(
                density=self.density, vel_x=self.vel_x, vel_y=self.vel_y, f_eq=self.f_eq
            )
            return
        roh = self.density
        u_x, u_y = self.velocity

        if self.lattice.eq_order == 2:
            self.f_eq[:, :, 0] = self.weights[0] * roh * (1 - 1.5 * (u_x**2 + u_y**2))
            self.f_eq[:, :, 1] = (
                    self.weights[1] * roh * (-1.5 * (u_x**2 + u_y**2) - 3 * u_y + 4.5 * u_y**2 + 1)
            )
            self.f_eq[:, :, 2] = (
                    self.weights[2] * roh * (-1.5 * (u_x**2 + u_y**2) - 3 * u_x + 4.5 * u_x**2 + 1)
            )
            self.f_eq[:, :, 3] = (
                    self.weights[3] * roh * (-1.5 * (u_x**2 + u_y**2) + 3 * u_x + 4.5 * u_x**2 + 1)
            )
            self.f_eq[:, :, 4] = (
                    self.weights[4] * roh * (-1.5 * (u_x**2 + u_y**2) + 3 * u_y + 4.5 * u_y**2 + 1)
            )
            self.f_eq[:, :, 5] = (
                    self.weights[5]
                    * roh
                    * (
                            -1.5 * (u_x**2 + u_y**2)
                            + 9.0 * u_x * u_y
                            - 3 * u_x
                            + 4.5 * u_x**2
                            - 3 * u_y
                            + 4.5 * u_y**2
                            + 1
                    )
            )
            self.f_eq[:, :, 6] = (
                    self.weights[6]
                    * roh
                    * (
                            -1.5 * (u_x**2 + u_y**2)
                            - 9.0 * u_x * u_y
                            + 3 * u_x
                            + 4.5 * u_x**2
                            - 3 * u_y
                            + 4.5 * u_y**2
                            + 1
                    )
            )
            self.f_eq[:, :, 7] = (
                    self.weights[7]
                    * roh
                    * (
                            -1.5 * (u_x**2 + u_y**2)
                            - 9.0 * u_x*u_y
                            - 3 * u_x
                            + 4.5 * u_x**2
                            + 3 * u_y
                            + 4.5 * u_y**2
                            + 1
                    )
            )
            self.f_eq[:, :, 8] = (
                    self.weights[8]
                    * roh
                    * (
                            -1.5 * (u_x**2 + u_y**2)
                            + 9.0 * u_x * u_y
                            + 3 * u_x
                            + 4.5 * u_x**2
                            + 3 * u_y
                            + 4.5 * u_y**2
                            + 1
                    )
            )

        if self.q == 17:
            self.f_eq[:, :, 0] = self.density * self.weights[0] * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq))
            self.f_eq[:, :, 4] = (
                    self.density
                    * self.weights[4]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + u_y / self.c_s_sq + u_y**2 / (2 * self.c_s_sq ** 2) + (
                    -u_x**2 * u_y - u_y**3) / (2 * self.c_s_sq ** 2) + u_y**3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 2] = (
                    self.density
                    * self.weights[2]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) - u_x / self.c_s_sq + u_x**2 / (2 * self.c_s_sq ** 2) + (
                    u_x**3 + u_y**2 * u_x) / (2 * self.c_s_sq ** 2) - u_x**3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 3] = (
                    self.density
                    * self.weights[3]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + u_x / self.c_s_sq + u_x**2 / (2 * self.c_s_sq ** 2) + (
                    -u_x**3 - u_y**2 * u_x) / (2 * self.c_s_sq ** 2) + u_x**3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 1] = (
                    self.density
                    * self.weights[1]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) - u_y / self.c_s_sq + u_y**2 / (2 * self.c_s_sq ** 2) + (
                    u_x**2 * u_y + u_y**3) / (2 * self.c_s_sq ** 2) - u_y**3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 7] = (
                    self.density
                    * self.weights[7]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (-u_x + u_y) / self.c_s_sq + (-u_x + u_y) ** 2 / (
                    2 * self.c_s_sq ** 2) + (u_x**3 - u_x**2 * u_y - u_y**3 + u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (-u_x + u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 8] = (
                    self.density
                    * self.weights[8]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (u_x + u_y) / self.c_s_sq + (u_x + u_y) ** 2 / (
                    2 * self.c_s_sq ** 2) + (-u_x**3 - u_x**2 * u_y - u_y**3 - u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (u_x + u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 5] = (
                    self.density
                    * self.weights[5]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (-u_x - u_y) / self.c_s_sq + (-u_x - u_y) ** 2 / (
                    2 * self.c_s_sq ** 2) + (u_x**3 + u_x**2 * u_y + u_y**3 + u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (-u_x - u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 6] = (
                    self.density
                    * self.weights[6]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (u_x - u_y) / self.c_s_sq + (u_x - u_y) ** 2 / (
                    2 * self.c_s_sq ** 2) + (-u_x**3 + u_x**2 * u_y + u_y**3 - u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (u_x - u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 11] = (
                    self.density
                    * self.weights[11]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (-2 * u_x + 2 * u_y) / self.c_s_sq + (
                        -2 * u_x + 2 * u_y) ** 2 / (
                               2 * self.c_s_sq ** 2) + (
                                   2 * u_x**3 - 2 * u_x**2 * u_y - 2 * u_y**3 + 2 * u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (-2 * u_x + 2 * u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 12] = (
                    self.density
                    * self.weights[12]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (2 * u_x + 2 * u_y) / self.c_s_sq + (
                        2 * u_x + 2 * u_y) ** 2 / (
                               2 * self.c_s_sq ** 2) + (
                                   -2 * u_x**3 - 2 * u_x**2 * u_y - 2 * u_y**3 - 2 * u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (2 * u_x + 2 * u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 9] = (
                    self.density
                    * self.weights[9]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (-2 * u_x - 2 * u_y) / self.c_s_sq + (
                        -2 * u_x - 2 * u_y) ** 2 / (
                               2 * self.c_s_sq ** 2) + (
                                   2 * u_x**3 + 2 * u_x**2 * u_y + 2 * u_y**3 + 2 * u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (-2 * u_x - 2 * u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 10] = (
                    self.density
                    * self.weights[10]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + (2 * u_x - 2 * u_y) / self.c_s_sq + (
                        2 * u_x - 2 * u_y) ** 2 / (
                               2 * self.c_s_sq ** 2) + (
                                   -2 * u_x**3 + 2 * u_x**2 * u_y + 2 * u_y**3 - 2 * u_y**2 * u_x) / (
                               2 * self.c_s_sq ** 2) + (2 * u_x - 2 * u_y) ** 3 / (6 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 16] = (
                    self.density
                    * self.weights[16]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + 3 * u_y / self.c_s_sq + 9 * u_y**2 / (
                        2 * self.c_s_sq ** 2) + (
                               -3 * u_x**2 * u_y - 3 * u_y**3) / (2 * self.c_s_sq ** 2) + 9 * u_y**3 / (
                                   2 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 14] = (
                    self.density
                    * self.weights[14]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) - 3 * u_x / self.c_s_sq + 9 * u_x**2 / (
                        2 * self.c_s_sq ** 2) + (
                               3 * u_x**3 + 3 * u_y**2 * u_x) / (2 * self.c_s_sq ** 2) - 9 * u_x**3 / (
                                   2 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 15] = (
                    self.density
                    * self.weights[15]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) + 3 * u_x / self.c_s_sq + 9 * u_x**2 / (
                        2 * self.c_s_sq ** 2) + (
                               -3 * u_x**3 - 3 * u_y**2 * u_x) / (2 * self.c_s_sq ** 2) + 9 * u_x**3 / (
                                   2 * self.c_s_sq ** 3))
            )
            self.f_eq[:, :, 13] = (
                    self.density
                    * self.weights[13]
                    * (1 - (u_x**2 + u_y**2) / (2 * self.c_s_sq) - 3 * u_y / self.c_s_sq + 9 * u_y**2 / (
                        2 * self.c_s_sq ** 2) + (
                               3 * u_x**2 * u_y + 3 * u_y**3) / (2 * self.c_s_sq ** 2) - 9 * u_y**3 / (
                                   2 * self.c_s_sq ** 3))
            )

    def do_collision(self):
        if self.pystencils:
            data_in = np.copy(self.data)
            self._collision_kernel(data=data_in, f_eq=self.f_eq, out=self.data)
        else:
            self.data = self.data * (1 - self.omega) + self.f_eq * self.omega

    def do_propagation(self):
        tmp = np.copy(self.data)
        if self.pystencils:
            layers = self.c_ix.max()
            data_in = add_ghost_layer(self.data, layers)
            data = np.zeros_like(data_in)
            self._propagation_kernel(data_in=data_in, data=data)
            self.data = data[layers:-layers, layers:-layers, :]
        else:
            for i in range(self.q):
                # Switch from Coordinates to Matrix indices: x --> y, y --> -x
                self.data[:, :, i] = np.roll(
                    tmp[:, :, i], (-1 * self.c_iy[i], self.c_ix[i]), axis=(0, 1)
                )

    def time_step(self, **kwargs):
        """
        Perform the time step.
        According to the book the time step algorithm can be split into different steps
        """
        self.iteration += 1
        # If Info Level Debugging is activated
        if logger.level <= 20:
            data_copy = np.copy(self.data)

        # Step 1: Compute Macroscopic moments
        self.update_macroscopic_fields()

        # Step 2: Update equilibrium distribution
        self.update_f_equilibrium()

        # Step 3: Write Macroscopic fields to disk for visualisation and post processing
        pass

        # Step 4: Perform Collision
        self.do_collision()

        # Step 5: Perform Streaming
        self.do_propagation()

        if not self._hurry and logger.level <= 20:
            l2diff = l2norm(data_copy, self.data)
            logger.debug("L2 Difference in this iteration: {}".format(l2diff))
            if l2diff < 10e-10:
                logger.info("Simulation has converged")

    @timer
    def iterate(self):
        """Function that gets called by the matplotlib animation
        Returns the values that are to be printed
        """
        self.time_step()

        if not self._hurry and self.fill_mode == "taylor-green":
            # taylor green has an analytical solution so we can output the accuracy
            dens_an, u_x_an, u_y_an = taylor_green(self.data.shape, self.iteration, self.relaxation_time, self.c_s_sq, self.test)
            vel_x, vel_y = self.velocity
            self.accuracy = l2norm(
                    u_x_an ** 2 + u_y_an ** 2,
                    vel_x ** 2 + vel_y ** 2,
                )
            logger.info(
                "L2 Distance to analytical solution of taylor Green in iteration {}: {}".format(
                    self.iteration,
                    self.accuracy

                )
            )
        return self.plot_data(self.plot_type)
