import numpy as np
from lbmweights import Lattice
from pystencils.session import *
import time

class StabilityException(Exception):
    pass


class LbmPs:
    """Implementation of the LBM using a single Kernel of Pystencils and lbmweights as initialisation scheme.

    Created for comparability, therefore the initialisation is done using viscocity instead of the
    usual relaxation Parameter. """
    def __init__(self, lattice_name, viscocity, resolution, iterations, u_0, calc_error=False, l2="squared", error_type="ux", check_stability=False, silent=False):
        """
        :param lattice_name: Name of the lattice for lbmweights
        :param tau: relaxation time
        :param resolution: resolution in each dimension
        :param iterations: amount of iterations if LbmPs.run_lbm() is called
        :param u_0: Maximum velocity for this initialization
        :param calc_error: True or False. If True, significantly slower, but error is calculated in self.errors
        """
        self.lattice_name = lattice_name
        self.lattice = Lattice.from_name(self.lattice_name)

        self.tau = float(viscocity/self.lattice.c_s_sq + 0.5)
        self.size = (resolution, resolution)
        self.iterations = iterations
        self.u_0 = u_0
        self.calc_error = calc_error
        self.check_stability = check_stability
        self.silent = silent

        self.times = [0, ]
        self.errors = [0, ]

        assignment_collection = self.lattice.assignment_collection
        self.upd_mac = self.macroscopic_moments_kernel(self.lattice)

        self.dh = ps.create_data_handling(self.size, periodicity=(True, True))
        src = self.dh.add_array('src', values_per_cell=self.lattice.q, ghost_layers=self.lattice.necessary_ghost_layers)
        dst = self.dh.add_array_like('dst', 'src')
        self.init_ker = self.init_kernel(self.lattice)
        self.copy_func = self.dh.synchronization_function(['src'], stencil='D2Q9')
        self.kernel = ps.create_kernel(assignment_collection).compile()
        self.initialize(self.dh)
        self.l2 = l2
        self.error_type = error_type
        self.reynold = self.u_0 * resolution / viscocity

    def cleanup(self, u_0, viscocity):
        """cleanup so kernels dont have to be calculated again"""
        self.u_0 = u_0
        self.tau = float(viscocity/self.lattice.c_s_sq + 0.5)
        self.times = [0, ]
        self.errors = [0, ]
        self.initialize(self.dh)
        self.reynold = self.u_0 * self.size[0] / viscocity

    def l2norm(self, approximation, solution):
        """
        Calculating the l2norm of two matrices.
        The matrices have to have the same shape

        :return Value that indicates how close mat2 i
        """
        assert (
                approximation.shape == solution.shape
        ), "The shape of the matrices have to be the same"

        diff = ((approximation - solution) ** 2).sum()
        abs = (solution ** 2).sum()
        norm = diff / abs
        if self.l2 != "squared":
            norm = np.sqrt(norm)
        return norm

    def taylor_green(self, t):
        N_x = self.size[0]
        N_y = self.size[1]
        u_0 = self.u_0
        c_s_sq = float(self.lattice.c_s_sq)
        p_0 = c_s_sq
        roh_avg = 0

        # Reynolds configs
        viscocity = c_s_sq * (self.tau - 0.5)

        kx = 2 * np.pi
        ky = 2 * np.pi

        assert abs(c_s_sq * (self.tau - 0.5) - viscocity) <= 1e-10

        t_d = 1 / (viscocity * ((2 * np.pi / N_x) ** 2 + (2 * np.pi / N_y) ** 2))
        if t == 0 and not self.silent:
            print("Viscocity: {}".format(viscocity))
            print("Relaxation: {}".format(self.tau))
            print("t_d: {}".format(t_d))
            print("Reynolds: {}".format(u_0*N_x/viscocity))
            print("Mach: {}".format(u_0/np.sqrt(c_s_sq)))


        X, Y = np.meshgrid(
            np.linspace(0, 1, num=N_x, endpoint=True),
            np.linspace(0, 1, num=N_y, endpoint=True),
            indexing='ij'
        )

        u_x = u_0 * (-1 * np.cos(kx * X) * np.sin(ky * Y)) * np.exp(- t / t_d)
        u_y = u_0 * (np.sin(kx * X) * np.cos(ky * Y)) * np.exp(- t / t_d)
        rho = (p_0 - roh_avg * u_0 ** 2 / 4 * (np.cos(2 * kx * X) + np.cos(2 * ky * Y)) * np.exp(-2 * t / t_d)) / c_s_sq

        return rho, u_x, u_y

    def macroscopic_moments_kernel(self, lattice):
        c_s_sq, weights, velocities = lattice.velocity_set()
        Q = lattice.q
        velocities = lattice.velocities

        c_ix = [vel[0] for vel in velocities]
        c_iy = [vel[1] for vel in velocities]

        src = ps.fields("src({Q}): double[2D]".format(Q=Q))
        velocity = ps.fields("velocity(2): [2D]")
        density, = sp.symbols("density, ")

        density_formula = sum([src[0, 0](i) for i in range(Q)])
        vel_x_formula = (sum([src[0, 0](i) * c_ix[i] for i in range(Q)])) / density
        vel_y_formula = (sum([src[0, 0](i) * c_iy[i] for i in range(Q)])) / density

        symbolic_description = [ps.Assignment(density, density_formula),
                                ps.Assignment(velocity[0, 0](0), vel_x_formula),
                                ps.Assignment(velocity[0, 0](1), vel_y_formula),
                                ]
        return ps.create_kernel(symbolic_description).compile()

    def init_kernel(self, lattice):
        c_s_sq, weights, velocities = lattice.velocity_set()
        Q = lattice.q
        velocities = lattice.velocities
        weights = [complex(x).real for x in weights]
        c_ix = [vel[0] for vel in velocities]
        c_iy = [vel[1] for vel in velocities]

        density = ps.fields("density: double[2D]")
        u_x = ps.fields("u_x: double[2D]")
        u_y = ps.fields("u_y: double[2D]")
        dst = ps.fields("dst({Q}): double[2D]".format(Q=Q))

        sym = []
        for i in range(Q):
            feq_formula = weights[i] * density[0, 0] * (
                    1
                    + (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) / c_s_sq
                    + (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 2 / (2 * c_s_sq ** 2)
                    - (u_x[0, 0] ** 2 + u_y[0, 0] ** 2) / (2 * c_s_sq)
            )

            if lattice.eq_order >= 3:
                feq_formula += weights[i] * density[0, 0] * (
                        -1 / (2 * c_s_sq ** 2) * (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) * (
                            u_x[0, 0] ** 2 + u_y[0, 0] ** 2)
                        + 4 / 3 * (1 / (2 * c_s_sq)) ** 3 * (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 3
                )
            if lattice.eq_order >= 4:
                feq_formula += weights[i] * density[0, 0] * (1 / (6 * c_s_sq ** 2) * (
                        (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 4 / (4 * c_s_sq ** 2) - (
                        3 * (u_x[0, 0] ** 2 + u_y[0, 0] ** 2) *
                        (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 2) / (2 * c_s_sq) + 3/4 * (
                                u_x[0, 0] ** 4 + u_y[0, 0] ** 4)))

            sym.append(ps.Assignment(dst[0, 0](i), feq_formula))
        return ps.create_kernel(sym).compile()

    def get_error(self, data, time):
        rho, u_x, u_y = self.taylor_green(time)
        velocities = np.zeros(self.size + (2,), dtype=np.float64)
        self.upd_mac(src=np.copy(data), velocity=velocities)

        if self.error_type == "ux":
            ana = u_x
            num = velocities[:, :, 0]
        elif self.error_type == "abs":
            ana = u_x ** 2 + u_y ** 2
            num = velocities[:, :, 0] ** 2 + velocities[:, :, 1] ** 2
        elif self.error_type == "density":
            ana = rho
            num = data.sum(axis=-1)
        else:
            raise ValueError("Error Type not implemented")

        return self.l2norm(num, ana)

    def vel_length(self, field):
        velocities = np.zeros(self.size + (2,), dtype=np.float64)
        self.upd_mac(src=np.copy(field), velocity=velocities)
        abs_vel = velocities[:, :, 0] ** 2 + velocities[:, :, 1] ** 2
        return abs_vel

    def initialize(self, dh):
        rho, u_x, u_y = self.taylor_green(0)

        entries_per_cell = dh.gather_array("src").shape[-1]
        init_feq = np.zeros((1,) + self.size + (entries_per_cell,))
        self.init_ker(density=rho, u_x=u_x, u_y=u_y, dst=init_feq[0])

        for block in dh.iterate(ghost_layers=False):
            x, y = block.cell_index_arrays
            block["src"][True] = init_feq
            pdfAC = dh.gather_array("src", ghost_layers=False)

    def run_lbm(self):
        """run this LatticeBoltzmannMethod"""
        timing = time.time()
        for i in np.arange(1, self.iterations + 1):
            self.copy_func()
            self.dh.run_kernel(self.kernel, omega=1/self.tau)
            self.dh.swap('src', 'dst')

            # Calculate error each iteration
            if self.calc_error:
                self.errors.append(self.get_error(self.dh.gather_array("src", ghost_layers=False), time=i))
                self.times.append(i)

            # Get idea of stability
            if self.check_stability:
                if self.dh.gather_array("src").sum(axis=2).min() < 0:
                    raise StabilityException("Negative population")




        total_time = time.time() - timing
        return total_time / self.iterations

