import pystencils as ps
from pystencils.session import *
import sympy as sp
import numpy as np
from utils.mylog import logger


def update_density_kernel(Q):
    """
    Q: Amount of stencils

    Returns a function that can be used to calculate the density for a given data point
        OUT = IN.sum(axis=2)
        ret(data=IN, density=OUT)

    """
    data = ps.fields("data({Q}) : double[2D]".format(Q=Q))
    density = ps.fields("density: [2D]")

    density_formula = sum([data[0, 0](i) for i in range(Q)])
    symbolic_description = [ps.Assignment(density[0, 0], density_formula)]

    kernel = ps.create_kernel(symbolic_description).compile()
    return kernel


def update_velocity_kernel(Q, c_ix, c_iy):
    """
    Get the compiled kernel for the velocity calculation for a given amount of stencils Q.

    """

    data = ps.fields("data({Q}) : double[2D]".format(Q=Q))
    density = ps.fields("density: [2D]")
    velocity = ps.fields("velocity(2): [2D]")

    vel_x_formula = (sum([data[0, 0](i) * c_ix[i] for i in range(Q)])) / density[0, 0]
    vel_y_formula = (sum([data[0, 0](i) * c_iy[i] for i in range(Q)])) / density[0, 0]
    symbolic_description = [
        ps.Assignment(velocity[0, 0](0), vel_x_formula),
        ps.Assignment(velocity[0, 0](1), vel_y_formula),
    ]

    kernel = ps.create_kernel(symbolic_description).compile()
    return kernel


def update_f_equilibrium_kernel(Q, weights, c_ix, c_iy, c_s_sq=1/3, test=False):
    f_eq = ps.fields("f_eq({Q}) : double[2D]".format(Q=Q))
    density = ps.fields("density: [2D]")
    vel_x = ps.fields("vel_x: [2D]")
    vel_y = ps.fields("vel_y: [2D]")
    u_x_sq, u_y_sq, u_x_u_y, u_sq, u_x, u_y = sp.symbols(
        "u_x_sq, u_y_sq, u_x_u_y, u_sq, u_x, u_y"
    )

    symbolic_description = [
        ps.Assignment(
            u_x, vel_x[0, 0]
        ),  # this is not necessary but more consistent. Sho
        ps.Assignment(u_y, vel_y[0, 0]),
    ]

    for i in range(Q):
        formula = weights[i] * density[0, 0] * (
                    1
                    + (u_x * c_ix[i] + u_y * c_iy[i]) / c_s_sq
                    + (u_x * c_ix[i] + u_y * c_iy[i]) ** 2 / (2 * c_s_sq ** 2)
                    - (u_x ** 2 + u_y ** 2) / (2 * c_s_sq)
                )
        if len(c_ix) >= 17:
            # 3rd? order macroscopic moments conserved
            if test:
                formula += weights[i] * density[0, 0] * (
                    -2 * (1/(2*c_s_sq))**2 * (c_ix[i] * u_x**3 + c_ix[i] * u_x * u_y**2 + c_iy[i] * u_y * u_x**2 + c_iy[i] * u_y**3)
                    + 4/3 * (1/(2*c_s_sq))**3 * (u_x * c_ix[i] + u_y * c_iy[i])**3
                )
            else:
                formula += weights[i] * density[0, 0] * (
                        -1 / (2 * c_s_sq ** 2) * (u_x * c_ix[i] + u_y * c_iy[i]) * (u_x ** 2 + u_y ** 2)
                        + 4 / 3 * (1 / (2 * c_s_sq)) ** 3 * (u_x * c_ix[i] + u_y * c_iy[i]) ** 3
                )

        if len(c_ix) >= 37:
            # According to Philippi this v37 is the smallest that has to conserve these.
            formula += weights[i] * density[0, 0] * \
                       (1/(6*c_s_sq**2) * (
                                          (u_x * c_ix[i] + u_y * c_iy[i])**4 / (4 * c_s_sq**2)  - (3 * (u_x**2 + u_y**2) * (
                                           u_x * c_ix[i] + u_y * c_iy[i])**2 )/ (2 * c_s_sq) + 3/4 * (u_x**4 + u_y**4)
            )
            )

        symbolic_description.append(
            ps.Assignment(
                f_eq[0, 0](i),
                formula,
            )
        )

    kernel = ps.create_kernel(symbolic_description).compile()
    return kernel


def collision_kernel(Q, omega):
    data = ps.fields("data({Q}): double[2D]".format(Q=Q))
    f_eq = ps.fields("f_eq({Q}): double[2D]".format(Q=Q))
    om = sp.symbols("om")

    out = ps.fields("out({Q}): double[2D]".format(Q=Q))
    symbolic_description = []
    for i in range(Q):
        symbolic_description.append(
            ps.Assignment(
                out[0, 0](i),
                (data[0, 0](i) * (1 - om) + f_eq[0, 0](i) * om).subs(om, omega),
            )
        )
    return ps.create_kernel(symbolic_description).compile()


def propagation_kernel(size, Q, c_ix, c_iy):
    data_in = ps.fields("data_in({Q}): double[2D]".format(Q=Q))
    data = ps.fields("data({Q}): double[2D]".format(Q=Q))

    symbolic_description = []
    for i in range(Q):
        """The way I use c_ix, c_iy might be confusing at first, but this is caused by the way numpy/pystencils works.
        For my stencil velocity definitions I used the usual mathematical convention of x as the horizontal coordinate 
        and y as the vertical. 
        Numpy and Pystencils use the Matrix convention, that means the first index indicates the vertical (y) and the 
        second the horizontal (x) coordinate. Additionally x is bigger, if it goes downwards. Hence the following 
        staement."""
        symbolic_description.append(
            ps.Assignment(data[0, 0](i), data_in[c_iy[i], -1 * c_ix[i]](i))
        )
    return ps.create_kernel(symbolic_description).compile()


def time_step_kernel(size, omega, lattice):
    c_s_sq, weights, velocity = lattice.velocity_set()
    c_ix = velocity[0]
    c_iy = velocity[1]
    Q = len(weights)
    ghost_layers = c_ix.max()

    # Data handling four ghost layers
    dh = ps.create_data_handling(domain_size=size, periodicity=(True, True))
    src = dh.add_array('src', values_per_cell=Q, ghost_layers=ghost_layers)
    dst = dh.add_array_like('dst', 'src')
    # Copy func used to copy over periodic values at boundaries. Only the space filling property of 'stencil'
    # is important
    copy_func = dh.synchronization_function(['src'], stencil='D2Q9')

    density, u_x, u_y = sp.symbols("density, u_x, u_y")

    density_formula = sum([src[0, 0](i) for i in range(Q)])
    vel_x_formula = (sum([src[0, 0](i) * c_ix[i] for i in range(Q)])) / density
    vel_y_formula = (sum([src[0, 0](i) * c_iy[i] for i in range(Q)])) / density

    feq_formula = [0] * Q
    for i in range(Q):
        feq_formula[i] = weights[i] * density * (
                1
                + (u_x * c_ix[i] + u_y * c_iy[i]) / c_s_sq
                + (u_x * c_ix[i] + u_y * c_iy[i]) ** 2 / (2 * c_s_sq ** 2)
                - (u_x ** 2 + u_y ** 2) / (2 * c_s_sq)
        )
        if len(c_ix) >= 17:
            feq_formula[i] += weights[i] * density * (
                    -2 * (1 / (2 * c_s_sq)) ** 2 * (
                        c_ix[i] * u_x ** 3 + c_ix[i] * u_x * u_y ** 2 + c_iy[i] * u_y * u_x ** 2 + c_iy[
                    i] * u_y ** 3)
                    + 4 / 3 * (1 / (2 * c_s_sq)) ** 3 * (u_x * c_ix[i] + u_y * c_iy[i]) ** 3
            )
        if len(c_ix) >= 37:
            feq_formula[i] += weights[i] * density * \
                           (1 / (6 * c_s_sq ** 2) * (
                                   (u_x * c_ix[i] + u_y * c_iy[i]) ** 4 / (4 * c_s_sq ** 2) - (
                                       3 * (u_x ** 2 + u_y ** 2) *
                                       (u_x * c_ix[i] + u_y * c_iy[i]) ** 2) / (2 * c_s_sq) + 3/4 * (
                                               u_x ** 4 + u_y ** 4)))

    # Update macroscopic moments
    symbolic_description = [ps.Assignment(density, density_formula),
                            ps.Assignment(u_x, vel_x_formula),
                            ps.Assignment(u_y, vel_y_formula), ]

    # Collision + Propagation
    for i in range(Q):
        symbolic_description.append(
            ps.Assignment(dst[-1 * c_iy[i], c_ix[i]](i), omega * feq_formula[i] + (1 - omega) * src[0, 0](i))
        )
    kernel = ps.create_kernel(symbolic_description).compile()

    return dh, copy_func, kernel


