import unittest
import logging
import time
import numpy as np
from unittest.mock import MagicMock
from lbm import LBM
from utils.mylog import logger
from utils.utils import l2norm
from lbmweights import Lattice


class TestLBM(unittest.TestCase):
    def setUp(self):
        size = (3, 3)
        self.lbm = LBM(size, Q=9, omega=1.2, pystencils=False)

        mock_fill = MagicMock()
        mock_fill.side_effect = self.mock_fill_side_effect
        self.lbm.fill = mock_fill
        self.lbm.fill()

    def mock_fill_side_effect(self):
        for j in range(3):
            for i in range(3):
                self.lbm.data[i, j, :] = np.ones((9)) * 1 / 9 * (i * 3 + j)

    def test_density(self):
        self.lbm.update_density()
        self.assertTrue(
            abs((self.lbm.density.reshape((9)) - np.arange(9))).max() <= 1e-10
        )

    def test_velocity(self):
        self.lbm.data[:, :] = np.arange(1, 10)
        self.lbm.update_density()
        self.lbm.update_velocity()
        vel_x, vel_y = self.lbm.velocity
        self.assertTrue(
            abs(vel_x[0, 0] - 0.13333333) <= 0.001 and abs(vel_y[0, 0] + 0.4) <= 0.001
        )

    def test_equilibrium(self):
        self.lbm.data[:, :] = np.arange(1, 10)
        solution = np.array(
            [
                0.51666667,
                1.26666667,
                0.31666667,
                2.06666667,
                14.66666667,
                6.06666667,
                2.31666667,
                13.26666667,
                4.51666667,
            ]
        )
        self.lbm.update_density()
        self.lbm.update_velocity()
        self.lbm.update_f_equilibrium()
        self.assertTrue((solution - self.lbm.f_eq[0, 0]).max() <= 0.001)

    def test_collision(self):
        # omega = 1.2
        self.lbm.data[:, :] = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
        self.lbm.update_macroscopic_fields()
        self.lbm.update_f_equilibrium()
        self.lbm.do_collision()
        self.assertTrue(
            (
                np.array([0.42, 1.12, -0.22, 1.68, 16.60, 6.08, 1.38, 14.32, 3.62])
                - self.lbm.data[0, 0]
            ).max()
            <= 0.001
        )

    def test_stream(self):
        # ignore collision to not change our density
        self.lbm.data = np.array([[x] * 9 for x in np.arange(0, 3 * 3)]).reshape(
            3, 3, 9
        )
        solution = np.arange(8, -1, -1)
        self.lbm.do_propagation()
        self.assertTrue((self.lbm.data[1, 1] == solution).all())


class TestLBM17(unittest.TestCase):
    def setUp(self):
        size = (7, 7)
        self.lbm = LBM(size, pystencils=False, Q=17)

        mock_fill = MagicMock()
        mock_fill.side_effect = self.mock_fill_side_effect
        self.lbm.fill = mock_fill
        self.lbm.fill()

    def mock_fill_side_effect(self):
        size = 7
        for j in range(size):
            for i in range(size):
                self.lbm.data[i, j, :] = np.ones((17)) * (i * size + j)

    def test_stream(self):
        # ignore collision to not change our density
        # self.lbm.data = test_dist
        self.lbm.do_propagation()
        # self.assertTrue((self.lbm.data[3:-3, 3:-3] == test_dist_solution[3:-3, 3:-3]).all())
        self.assertTrue(
            (
                self.lbm.data[3, 3]
                == np.array(
                    [24, 31, 25, 23, 17, 32, 30, 18, 16, 40, 36, 12, 8, 45, 27, 21, 3]
                )
            ).all()
        )
        self.assertTrue(
            (
                self.lbm.data[0, 0]
                == np.array(
                    [0, 7, 1, 6, 42, 8, 13, 43, 48, 16, 19, 37, 40, 21, 3, 4, 28]
                )
            ).all()
        )
        self.assertTrue(
            (
                self.lbm.data.min(axis=2)
                == [
                    [0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 0.0],
                    [2.0, 3.0, 0.0, 1.0, 2.0, 0.0, 1.0],
                    [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                    [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                    [2.0, 3.0, 0.0, 1.0, 2.0, 0.0, 1.0],
                    [0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 0.0],
                ]
            ).all()
        )


class TestV17Updates(unittest.TestCase):
    """
    Test if recent changes have resulted in significant changes in both time and value.
    """

    def setUp(self):
        logger.setLevel(logging.WARNING)
        size = (7, 7)
        lattice = Lattice.from_name("D2V17")
        self.lbm = LBM(
            size,
            pystencils=False,
            lattice=lattice,
            omega=0.4,
            plot_type="density",
            fill_mode="taylor-green",
            test=False,
        )
        self.lbmt = LBM(
            size,
            pystencils=False,
            lattice=lattice,
            omega=0.4,
            plot_type="density",
            fill_mode="taylor-green",
            test=True,
        )

    def test_equal(self):
        # test equality
        self.assertTrue((self.lbm.density == self.lbmt.density).all())
        self.assertTrue((self.lbm.vel_x == self.lbmt.vel_x).all())
        self.assertTrue((self.lbm.vel_y == self.lbmt.vel_y).all())

        iterations = 1000
        t0 = time.time()
        for i in range(iterations):
            self.lbm.iterate()
        t1 = time.time()
        for i in range(iterations):
            self.lbmt.iterate()
        t2 = time.time()
        logger.setLevel(logging.INFO)
        logger.info(
            "Execution time of {} iterations of test==false:  {} vs test==true implementation {}".format(
                iterations, t1 - t0, t2 - t1
            )
        )
        logger.setLevel(logging.WARNING)

        self.assertTrue((abs(self.lbm.density - self.lbmt.density)).max() < 1e-15)
        self.assertTrue((abs(self.lbm.vel_x - self.lbmt.vel_x)).max() < 1e-15)
        self.assertTrue((abs(self.lbm.vel_y - self.lbmt.vel_y)).max() < 1e-15)


class TestQ9Updates(unittest.TestCase):
    """
    Test if recent changes have resulted in significant changes in both time and value.
    """

    def setUp(self):
        logger.setLevel(logging.WARNING)
        size = (7, 7)
        lattice = Lattice.from_name("D2Q9")
        self.lbm = LBM(
            size,
            pystencils=False,
            lattice=lattice,
            omega=0.4,
            plot_type="density",
            fill_mode="taylor-green",
            test=False,
        )
        self.lbmt = LBM(
            size,
            pystencils=False,
            lattice=lattice,
            omega=0.4,
            plot_type="density",
            fill_mode="taylor-green",
            test=True,
        )

    def test_equal(self):
        # test equality
        self.assertTrue((self.lbm.density == self.lbmt.density).all())
        self.assertTrue((self.lbm.vel_x == self.lbmt.vel_x).all())
        self.assertTrue((self.lbm.vel_y == self.lbmt.vel_y).all())

        iterations = 1000
        t0 = time.time()
        for i in range(iterations):
            self.lbm.iterate()
        t1 = time.time()
        for i in range(iterations):
            self.lbmt.iterate()
        t2 = time.time()
        logger.setLevel(logging.INFO)
        logger.info(
            "Execution time of {} iterations of test==false:  {} vs test==true implementation {}".format(
                iterations, t1 - t0, t2 - t1
            )
        )
        logger.setLevel(logging.WARNING)

        self.assertTrue((abs(self.lbm.density - self.lbmt.density)).max() < 1e-15)
        self.assertTrue((abs(self.lbm.vel_x - self.lbmt.vel_x)).max() < 1e-15)
        self.assertTrue((abs(self.lbm.vel_y - self.lbmt.vel_y)).max() < 1e-15)


class TestFeqUpdates(unittest.TestCase):
    def setUp(self):
        logger.setLevel(logging.WARNING)
        size = (7, 7)
        lattice = Lattice.from_name("D2V17")
        self.lbmFalse = LBM(
            size,
            pystencils=False,
            lattice=lattice,
            omega=0.4,
            plot_type="density",
            fill_mode="taylor-green",
            test=False,
        )
        self.lbmTrue = LBM(
            size,
            pystencils=False,
            lattice=lattice,
            omega=0.4,
            plot_type="density",
            fill_mode="taylor-green",
            test=True,
        )

        self.lbmFalse.update_macroscopic_fields()
        self.lbmTrue.update_macroscopic_fields()

        self.lbmFalse.density = np.arange(0, 7*7).reshape((7,7))
        self.lbmTrue.density = np.arange(0, 7*7).reshape((7,7))

        self.lbmFalse.vel_x = 3 * np.arange(0, 7*7).reshape((7,7))
        self.lbmFalse.vel_y = 0.3 * np.arange(0, 7*7).reshape((7,7))

        self.lbmTrue.vel_x = 3 * np.arange(0, 7*7).reshape((7,7))
        self.lbmTrue.vel_y = 0.3 * np.arange(0, 7*7).reshape((7,7))

    def test_f_equilibrium(self):
        print("-"*60 + "Test f equilibrium in Feq Updates")
        assert (self.lbmTrue.test != self.lbmFalse.test)
        assert (self.lbmTrue.shape == self.lbmFalse.shape)

        assert (self.lbmFalse.density == self.lbmTrue.density).all()
        assert (self.lbmFalse.vel_x == self.lbmTrue.vel_x).all()
        assert (self.lbmFalse.vel_y == self.lbmTrue.vel_y).all()

        self.lbmTrue.update_f_equilibrium()
        self.lbmFalse.update_f_equilibrium()

        shape = self.lbmTrue.shape
        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(shape[2]):
                    try:
                        self.assertAlmostEqual(self.lbmTrue.f_eq[i, j, k], self.lbmFalse.f_eq[i, j, k], places=6)
                    except AssertionError as e:
                        logger.warning("Index = x={} y={} velocity = {}".format(i, j, k))
                        raise e

        max_diff = (abs(self.lbmTrue.f_eq.flatten()) - abs(self.lbmFalse.f_eq.flatten())).max()
        if max_diff > 1e-17:
            logger.warning("!" * 20 + "NOT EXACTLY : {}".format(max_diff))


if __name__ == "__main__":
    unittest.main()
