import unittest
import numpy as np
from lbm import LBM


class TestPystencils9(unittest.TestCase):
    def setUp(self):
        size = (7, 7)
        self.lbm = LBM(size, Q=9, omega=0.4, pystencils=False)
        self.lbmps = LBM(size, Q=9, omega=0.4, pystencils=True)

        self.lbm.data = np.arange(0, 7 * 7 * 9, dtype=np.float64).reshape(7, 7, 9)
        self.lbmps.data = np.arange(0, 7 * 7 * 9, dtype=np.float64).reshape(7, 7, 9)

    def test_density(self):
        self.lbm.update_density()
        self.lbmps.update_density()
        self.assertTrue((self.lbm.density == self.lbmps.density).all())

    def test_velocity(self):
        self.lbm.update_density()
        self.lbmps.update_density()

        self.lbm.update_velocity()
        self.lbmps.update_velocity()

        self.assertTrue((self.lbm.vel_x == self.lbmps.vel_x).all())
        self.assertTrue((self.lbm.vel_y == self.lbmps.vel_y).all())

    def test_f_equilibrium(self):
        self.lbm.update_density()
        self.lbmps.update_density()

        self.lbm.update_velocity()
        self.lbmps.update_velocity()

        self.lbm.update_f_equilibrium()
        self.lbmps.update_f_equilibrium()

        self.assertTrue(
            (
                (self.lbm.f_eq - self.lbmps.f_eq)
                <= (np.ones_like(self.lbmps.f_eq) * 1e-12)
            ).all()
        )

    def test_collision(self):
        f_eq = np.random.rand(7, 7, 9)
        self.lbm.f_eq = np.copy(f_eq)
        self.lbmps.f_eq = np.copy(f_eq)

        self.lbm.do_collision()
        self.lbmps.do_collision()

        self.assertTrue((self.lbm.data == self.lbmps.data).all())

    def test_propagation(self):
        data = np.zeros_like(self.lbm.data)
        for row in range(7):
            for col in range(7):
                for stencil in range(9):
                    data[row, col, stencil] = row * 10 + col + (stencil + 1) * 0.1

        self.lbm.data = np.copy(data)
        self.lbmps.data = np.copy(data)

        self.lbm.do_propagation()
        self.lbmps.do_propagation()
        self.assertTrue((self.lbm.data == self.lbmps.data).all())
        self.assertTrue((self.lbm.data == self.lbmps.data).all())


class TestPystencils17(unittest.TestCase):
    def setUp(self):
        size = (7, 7)
        self.lbm = LBM(size, Q=17, omega=1.5, pystencils=False)
        self.lbmps = LBM(size, Q=17, omega=1.5, pystencils=True)

        self.lbm.data = np.arange(0, 7 * 7 * 17, dtype=np.float64).reshape(7, 7, 17)
        self.lbmps.data = np.arange(0, 7 * 7 * 17, dtype=np.float64).reshape(7, 7, 17)

    def test_density(self):
        self.lbm.update_density()
        self.lbmps.update_density()
        self.assertTrue((self.lbm.density == self.lbmps.density).all())

    def test_velocity(self):
        self.lbm.update_density()
        self.lbmps.update_density()

        self.lbm.update_velocity()
        self.lbmps.update_velocity()

        self.assertTrue((self.lbm.data == self.lbmps.data).all())
        self.assertTrue((self.lbm.density == self.lbmps.density).all())
        self.assertTrue((self.lbm.vel_x == self.lbmps.vel_x).all())
        self.assertTrue((self.lbm.vel_y == self.lbmps.vel_y).all())

    def test_f_equilibrium(self):
        self.lbm.update_density()
        self.lbmps.update_density()

        self.lbm.update_velocity()
        self.lbmps.update_velocity()

        self.lbm.update_f_equilibrium()
        self.lbmps.update_f_equilibrium()

        self.assertTrue(
            (
                (self.lbm.f_eq - self.lbmps.f_eq)
                <= (np.ones_like(self.lbmps.f_eq) * 1e-12)
            ).all()
        )

    def test_collision(self):
        f_eq = np.random.rand(7, 7, 17)
        self.lbm.f_eq = np.copy(f_eq)
        self.lbmps.f_eq = np.copy(f_eq)

        self.lbm.do_collision()
        self.lbmps.do_collision()

        self.assertTrue((self.lbm.data == self.lbmps.data).all())

    def test_propagation(self):
        data = np.zeros_like(self.lbm.data)
        for row in range(7):
            for col in range(7):
                for stencil in range(17):
                    data[row, col, stencil] = row * 10 + col + (stencil + 1) * 0.01

        self.lbm.data = np.copy(data)
        self.lbmps.data = np.copy(data)

        self.lbm.do_propagation()
        self.lbmps.do_propagation()
        self.assertTrue((self.lbm.data == self.lbmps.data).all())
        self.assertTrue((self.lbm.data == self.lbmps.data).all())


if __name__ == "__main__":
    unittest.main()
