import time
import numpy as np
from utils.mylog import logger

pause = True
pause_immediately = False


class Timer:
    times = dict()

    @staticmethod
    def get_times():
        return Timer.times

    @staticmethod
    def purge_times():
        """Purges all time entries in the time dict. Useful to compare multiple implementations by calling
        them after another and purging the dict in between."""
        Timer.times = dict()


    @staticmethod
    def timer(func, *args, **kwargs):
        """
        Python decorator to time a function. Add this to a function to create a dict entry
        for this function that saves the best computation time.
        """

        def avg_time(func_name, time):
            if not Timer.times.get(func_name):
                Timer.times[func_name] = [1, time]
            else:
                avg = Timer.times[func_name][1]
                Timer.times[func_name][0] += 1
                x = Timer.times[func_name][0]
                Timer.times[func_name][1] = (x - 1) / x * avg + 1 / x * time
            return Timer.times[func_name][1]

        def best_time(func_name, time):
            if not Timer.times.get(func_name):
                Timer.times[func_name] = time
            else:
                if time < Timer.times[func_name]:
                    Timer.times[func_name] = time
            return Timer.times[func_name]

        def wrapper(*args, **kwargs):
            start_time = time.time()
            ret = func(*args, **kwargs)
            elapsed_time = time.time() - start_time

            logger.info(
                " - Execution time of {0}: {1}".format(
                    func.__name__, best_time(func.__name__, elapsed_time)
                )
            )
            return ret

        return wrapper


def l2norm(approximation, solution):
    """
    Calculating the l2norm of two matrices.
    The matrices have to have the same shape

    :return Value that indicates how close mat2 i
    """
    assert (
        approximation.shape == solution.shape
    ), "The shape of the matrices have to be the same"

    diff = ((approximation - solution) ** 2).sum()
    abs = (solution ** 2).sum()
    norm = diff / abs
    return norm


def taylor_green(data_shape, t, tau, c_s_sq, test=False):
    N_x = data_shape[0]
    N_y = data_shape[1]
    u_0 = 0.04
    p_0 = c_s_sq
    roh_avg = 0.1
    # Reynolds configs
    viscocity = c_s_sq * (tau - 0.5)

    kx = 2 * np.pi
    ky = 2 * np.pi

    if t == 0:
        t_d = 1 / (viscocity * ((kx/N_x) ** 2 + (ky/N_y) ** 2))
        mach = u_0 / np.sqrt(c_s_sq)
        reynolds = u_0 * N_x / viscocity
        logger.info("Tau: {}".format(tau))
        logger.info("Viscosity: {}".format(viscocity))
        logger.info("Reynolds: {}".format(reynolds))
        logger.info("Mach: {}".format(mach))
        logger.info("Decay Time: {}".format(t_d))
    assert abs(c_s_sq * (tau - 0.5) - viscocity) <= 1e-10

    X, Y = np.meshgrid(
        np.linspace(0, 1, num=N_x, endpoint=True),
        np.linspace(0, 1, num=N_y, endpoint=True),
    )

    t_d = 1 / (viscocity * ((2 * np.pi/N_x) ** 2 + (2 * np.pi/N_y) ** 2))
    u_x = u_0 * (-1 * np.cos(kx * X) * np.sin(ky * Y)) * np.exp(- t / t_d)  # y in visualisation
    u_y = u_0 * (-1 * np.sin(kx * X) * np.cos(ky * Y)) * np.exp(- t / t_d)  # Minus for visualisation
    roh = (p_0 - roh_avg * u_0 ** 2 / 4 * (np.cos(2 * kx * X) + np.cos(2 * ky * Y)) * np.exp(-2*t/t_d))/c_s_sq

    return roh, u_x, u_y


def add_ghost_layer(data, layers):
    """
    :param data: 3 Dimenstional Numpy Array
    :param layers: Layers that are added on each side of the field
    :return: numpy array with increased size (2 * layers in the first two Dimensions)

    Example:
    NOTE: 0 := [...] some Array. Can't be an integer. Use data.sum(axis=2) for integers!
    layer = 1
    data = [[0 1 2]
            [3 4 5]
            [6 7 8]]

    return = [[8 6 7 8 6]
              [2 0 1 2 0]
              [5 3 4 5 3]
              [8 6 7 8 6]
              [2 0 1 2 0]]
    """
    rows = data.shape[0]
    stencils = data.shape[2]

    output = np.concatenate((data[-layers:, :, :], data, data[:layers, :, :]), axis=0)
    if layers == 1:
        left_side = np.concatenate(
            ([[data[-1, -1, :]]], data[:, -layers:, :], [[data[0, -1, :]]])
        ).reshape(rows + layers * 2, 1, stencils)
        right_side = np.concatenate(
            ([[data[-1, 0, :]]], data[:, :layers, :], [[data[0, 0, :]]])
        ).reshape(rows + layers * 2, 1, stencils)
    if layers != 1:
        left_side = np.concatenate(
            (
                data[-layers:, -layers:, :],
                data[:, -layers:, :],
                data[:layers, -layers:, :],
            )
        ).reshape(rows + layers * 2, layers, stencils)
        right_side = np.concatenate(
            (data[-layers:, :layers, :], data[:, :layers, :], data[:layers, :layers, :])
        ).reshape(rows + layers * 2, layers, stencils)
    output = np.concatenate((left_side, output, right_side), axis=1)
    return output
