"""Collection of a number of visualizations that may or may not end in the final thesis.

A function usually corresponds to one visualization and can be called in the main function to recapitulate
how the functions were created. In addition some visualization may require older versions
of the script. They can be found using the git versioning system. If a function provides a <hash> in it's
description this is what you do for the exact result:

git checkout <hash>
"""

# Generic imports

from lbmweights import Lattice
from lbm import LBM
from utils.utils import logger
import numpy as np
from fractions import Fraction
from utils.utils import Timer
import click
import time
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.animation as animation
import matplotlib.colors
import logging
from lbmps import LbmPs, StabilityException
from pprint import pprint


def time_compare(func1, func2, func3=None, iters=100):
    """Compare execution times of *iters* calls of two functions.
    Times are returned in order."""
    t_start = time.time()
    for i in range(iters):
        func1()
    t_break = time.time()
    for i in range(iters):
        func2()
    t_stop = time.time()

    f1_time = t_break - t_start
    f2_time = t_stop - t_break


    f1_time = t_break - t_start
    f2_time = t_stop - t_break
    if func3:
        for i in range(iters):
            func3()
        f3_time = time.time() - t_stop
        return f1_time, f2_time, f3_time

    return f1_time, f2_time


def compare_numpy_accuracy():
    """Compare the accuracy of the NumPy Implementations of D2Q9 and D2V17 for the Taylor Green Flow.

    hash: 4b4bec85ae071726dc81a39b20b5609d631f95cc
    """
    # lattices to be used
    lattice_q9 = Lattice.from_name("D2Q9")
    lattice_v17 = Lattice.from_name("D2V17")
    lattice_v17.c_s_sq

    # Simulation params
    s = 100
    size = (s, s)
    tau = 0.55
    iterations = 1000

    # c_s_sq different for both implementations therefore tau has to be changed so both simulate the same thing
    viscosity_q9 = 1 / 3 * (tau - 0.5)
    tau_v17 = viscosity_q9 / float(lattice_v17._c_s_sq) + 0.5
    omega = 1 / tau
    omegav17 = 1 / tau_v17
    lbm_q9 = LBM(size=size, lattice=lattice_q9, pystencils=False, omega=omega, fill_mode="taylor-green")
    lbm_v17 = LBM(size=size, lattice=lattice_v17, pystencils=False, omega=omegav17, fill_mode="taylor-green")

    accuracy_q9 = np.zeros((iterations))
    accuracy_v17 = np.zeros((iterations))
    for i in range(iterations):
        lbm_q9.iterate()
        accuracy_q9[i] = lbm_q9.accuracy
    for i in range(iterations):
        lbm_v17.iterate()
        accuracy_v17[i] = lbm_v17.accuracy

    # Plotting
    x = np.arange(0, iterations)
    plt.plot(x, accuracy_q9, 'r', label="Q9")
    plt.plot(x, accuracy_v17, 'b', label="V17")
    plt.legend()
    plt.show()


def compare_numpy_performance():
    """Chapter Benchmark Comparison:

    Compare V17 vs Q9 written using pure NumPy. Special Focus on f_equilibrium as the new
    bottleneck.
    hash: 11124ec4e92b9e08a5ee982393d595f21f0b4dde

    """
    lattice_q9 = Lattice.from_name("D2Q9")
    lattice_v17 = Lattice.from_name("D2V17")

    # Simulation params
    s = 100
    size = (s, s)
    tau = 0.8
    iterations = 100

    # c_s_sq different for both implementations therefore tau has to be changed so both simulate the same thing
    viscosity_q9 = 1 / 3 * (tau - 0.5)
    #tau_v17 = viscosity_q9 / float(lattice_v17._c_s_sq) + 0.5
    omega = 1 / tau
    #omegav17 = 1 / tau_v17

    lbm_q9 = LBM(size=size, lattice=lattice_q9, pystencils=False, omega=omega, fill_mode="taylor-green")
    lbm_v17 = LBM(size=size, lattice=lattice_q9, pystencils=False, omega=omega, fill_mode="taylor-green", test=True)

    # Whole algorithm
    import time
    t0 = time.time()
    for i in range(iterations):
        lbm_q9.iterate()
    t1 = time.time()
    for i in range(iterations):
        lbm_v17.iterate()
    t2 = time.time()
    print("Iterate v17: {} vs q9:{}".format(t2 - t1, t1 - t0))
    print("Iterate v17 takes {} times as long".format((t2 - t1) / (t1 - t0)))

    # update_f_equilibrium
    t0 = time.time()
    for i in range(iterations):
        lbm_q9.update_f_equilibrium()
    t1 = time.time()
    for i in range(iterations):
        lbm_v17.update_f_equilibrium()
    t2 = time.time()
    print("Equilibrium v17: {} vs q9:{}".format(t2 - t1, t1 - t0))
    print("Equilibrium v17 takes {} times as long".format((t2 - t1) / (t1 - t0)))


def compare_v37_pystencils():
    lattice_v17 = Lattice.from_name("D2V17")
    lattice_v37 = Lattice.from_name("D2V37")

    s = 100
    size = (s, s)
    tau = 0.8
    iterations = 100

    viscosity = 1 / 3 * (tau - 0.5)
    omega_v17 = 1 / (viscosity / float(lattice_v17.c_s_sq) + 0.5)
    omega_v37 = 1 / (viscosity / float(lattice_v37.c_s_sq) + 0.5)

    lbm_v17 = LBM(size=size, lattice=lattice_v17, pystencils=True, omega=omega_v17, fill_mode="taylor-green")
    lbm_v37 = LBM(size=size, lattice=lattice_v37, pystencils=True, omega=omega_v37, fill_mode="taylor-green")

    # Whole algorithm
    import time
    t0 = time.time()
    for i in range(iterations):
        lbm_v17.iterate()
    t1 = time.time()
    for i in range(iterations):
        lbm_v37.iterate()
    t2 = time.time()
    print("Iterate v37: {} vs v17:{}".format(t2 - t1, t1 - t0))
    print("Iterate v37 takes {} times as long".format((t2 - t1) / (t1 - t0)))

    # update_f_equilibrium
    t0 = time.time()
    for i in range(iterations):
        lbm_v17.update_f_equilibrium()
    t1 = time.time()
    for i in range(iterations):
        lbm_v37.update_f_equilibrium()
    t2 = time.time()
    print("Equilibrium v37: {} vs v17:{}".format(t2 - t1, t1 - t0))
    print("Equilibrium v37 takes {} times as long".format((t2 - t1) / (t1 - t0)))


def compare_v37_accuracy():
    # lattices to be used
    lattice_q9 = Lattice.from_name("D2Q9")
    lattice_v17 = Lattice.from_name("D2V17")
    lattice_v37 = Lattice.from_name("D2V37")

    # Simulation params
    s = 100
    size = (s, s)
    tau = 0.55
    iterations = 1000

    # c_s_sq different for both implementations therefore tau has to be changed so both simulate the same thing
    viscosity = 1 / 3 * (tau - 0.5)
    omega_q9 = 1 / tau
    omega_v17 = 1 / (viscosity / float(lattice_v17.c_s_sq) + 0.5)
    omega_v37 = 1 / (viscosity / float(lattice_v37.c_s_sq) + 0.5)

    lbm_q9 = LBM(size=size, lattice=lattice_q9, pystencils=True, omega=omega_q9, fill_mode="taylor-green")
    lbm_v17 = LBM(size=size, lattice=lattice_v17, pystencils=True, omega=omega_v17, fill_mode="taylor-green")
    lbm_v37 = LBM(size=size, lattice=lattice_v37, pystencils=True, omega=omega_v37, fill_mode="taylor-green")

    accuracy_q9 = np.zeros((iterations))
    accuracy_v17 = np.zeros((iterations))
    accuracy_v37 = np.zeros((iterations))

    for i in range(iterations):
        lbm_q9.iterate()
        accuracy_q9[i] = lbm_q9.accuracy
    for i in range(iterations):
        lbm_v17.iterate()
        accuracy_v17[i] = lbm_v17.accuracy
    for i in range(iterations):
        lbm_v37.iterate()
        accuracy_v37[i] = lbm_v37.accuracy

    # Plotting
    x = np.arange(0, iterations)
    plt.plot(x, accuracy_q9, 'r', label="Q9")
    plt.plot(x, accuracy_v17, 'b', label="V17")
    plt.plot(x, accuracy_v37, 'g', label="V37")
    plt.legend()
    plt.show()


def compare_pystencils_numpy(resolution, iterations):
    lattice_v17 = Lattice.from_name("D2V17")
    size = (resolution, resolution)

    lbm_np = LBM(size=size, lattice=lattice_v17, pystencils=False, omega=0.8, fill_mode="taylor-green")
    lbm_ps = LBM(size=size, lattice=lattice_v17, pystencils=True, omega=0.8, fill_mode="taylor-green")

    t_start = time.time()
    for i in range(iterations):
        lbm_np.iterate()
    t_break = time.time()
    for i in range(iterations):
        lbm_ps.iterate()
    t_stop = time.time()

    print("Pystencils = {} Numpy = {}".format((t_stop - t_break), (t_break - t_start)))


def time_step_kernel(resolution):
    lattice_q9 = Lattice.from_name("D2Q9")
    size = (resolution, resolution)
    omega = 0.8

    lbm_ps = LBM(size=size, lattice=lattice_q9, pystencils=True, omega=omega, fill_mode="taylor-green")

    import pystencils_kernels as pystencils_kernels
    dh, copy_func, kernel = pystencils_kernels.time_step_kernel(size=size, omega=omega, lattice=lattice_q9)
    dh.fill('src', 0, ghost_layers=True)
    dh.fill('dst', 0, ghost_layers=True)
    for block in dh.iterate(ghost_layers=False):
        x, y = block.cell_index_arrays
        block["src"][True] = lbm_ps.data[x, y]
    assert (dh.gather_array('src', ghost_layers=False) == lbm_ps.data).all()

    #print(dh.gather_array('src',ghost_layers=True)[resolution + 1, resolution + 1, :])
    copy_func()
    assert (dh.gather_array('src', ghost_layers=True)[0, 0] == dh.gather_array('src', ghost_layers=False)[-1, -1]).all()
    assert (dh.gather_array('src', ghost_layers=True)[3, 0] == dh.gather_array('src', ghost_layers=False)[2, -1]).all()
    #print(dh.gather_array('src',ghost_layers=True)[resolution + 1, resolution + 1, :])
    #print(abs(dh.gather_array('src', ghost_layers=True)[1, 1, :] - dh.gather_array('src',ghost_layers=True)[resolution + 1, resolution + 1, :]) < 1e-160)

    dh.run_kernel(kernel)
    dh.swap('src', 'dst')
    lbm_ps.iterate()

    tx, ty= 0, 0
    print(dh.gather_array('src', ghost_layers=False)[tx, ty, :])
    print(lbm_ps.data[tx, ty, :])
    print(abs((dh.gather_array('src', ghost_layers=False)[tx, ty, :] - lbm_ps.data[tx, ty, :])) <= 1e-10)
    print(abs(dh.gather_array('src', ghost_layers=False)[1:-2, 1:-2, :] - lbm_ps.data[1:-2, 1:-2, :]).max())
    #print((dh.gather_array('src', ghost_layers=False)[:, :, 3] == lbm_ps.data[:, :, 3]))

    assert (dh.gather_array('src', ghost_layers=False) == lbm_ps.data).all()


def performance_numpy_vs_pystencils_v17(average_over, iterations):
    """ Hash: 0aea168bae3a3f5ca652821fa769346a799a5344
    Get a barplot, as done in chapter Perforamance comparison, that compares the computation
    time of the V17 lattice for the Pystencils and the NumPy implementation.
    The average execution time and performance gain is plotted over *iteration* and over
    alls system sizes given in *average_over*.

    Example:
    performance_numpy_vs_pystencils_v17(average_over=[10, 50, 100, 200], iterations=100)
    Compute the avarage of 100 iterations for system size 10, 50, 100 and 200.

    :param average_over: list with integer values to be used as simulation size in both dimensions
    :param iterations: iterations to average over
    """
    lattice_v17 = Lattice.from_name("D2V17")
    omega = 0.8

    # ======== Compute the average computation time for some of
    avg_dict = dict()
    avg_time_dict = dict()
    for i, av_over in enumerate(average_over):
        # Create new instances
        lbm_np = LBM(size=(int(av_over), int(av_over)), lattice=lattice_v17, pystencils=False, omega=omega, fill_mode="taylor-green")
        lbm_ps = LBM(size=(int(av_over), int(av_over)), lattice=lattice_v17, pystencils=True, omega=omega, fill_mode="taylor-green")

        time_dict = dict()
        time_dict["time_step"] = list(time_compare(lbm_np.time_step, lbm_ps.time_step, iters=iterations))
        time_dict["update_f_equilibrium"] = list(time_compare(lbm_np.update_f_equilibrium, lbm_ps.update_f_equilibrium, iters=iterations))
        time_dict["update_macroscopic_fields"] = list(time_compare(lbm_np.update_macroscopic_fields, lbm_ps.update_macroscopic_fields, iters=iterations))
        time_dict["collision"] = list(time_compare(lbm_np.do_collision, lbm_ps.do_collision, iters=iterations))
        time_dict["propagation"] = list(time_compare(lbm_np.do_propagation, lbm_ps.do_propagation, iters=iterations))

        print("Performance improvement for {}".format((int(av_over), int(av_over))))
        for key, value in time_dict.items():
            improvement = value[0]/value[1]
            avg = avg_dict.get(key)
            avg_time = avg_time_dict.get(key)
            if not avg:
                avg_dict[key] = improvement
                avg_time_dict[key] = [value[0], value[1]]
            else:
                avg_dict[key] = i / (i+1) * avg + 1 / (i + 1) * improvement
                avg_time_dict[key][0] = i / (i+1) * avg_time[0] + 1 / (i + 1) * value[0]
                avg_time_dict[key][1] = i / (i+1) * avg_time[1] + 1 / (i + 1) * value[1]

            print("{}: Numpy {}s,  Pystencils: {}s, --- Improvement: {}".format(key, value[0], value[1], improvement))

        print("{} finished ---".format(i))

    # ======== Plot as bar chart
    import matplotlib.pyplot as plt; plt.rcdefaults()
    import numpy as np
    import matplotlib.pyplot as plt

    functions = ["update_macroscopic_fields", "update_f_equilibrium", "collision", "propagation", "time_step"]
    times = np.zeros((2, 5), dtype=np.float64)
    for i, func in enumerate(functions):
        times[0, i] = avg_time_dict[func][0]
        times[1, i] = avg_time_dict[func][1]

    normal = times[0, :]
    pystencils = times[1, :]

    fig, ax = plt.subplots()
    index = np.arange(len(pystencils))
    bar_width = 0.3
    opacity = 0.8

    rects1 = plt.barh(index + bar_width/2, normal, bar_width, alpha=opacity, color="b", label="NumPy")
    rects2 = plt.barh(index - bar_width/2, pystencils, bar_width, alpha=opacity, color="cyan", label="Pystencils")

    def autolabel(rects):
        """Label improvement next to bars"""
        for i, rect in enumerate(rects):
            ax.annotate('{0:.2f}'.format(avg_dict[functions[i]]),
                        xy=(rect.get_x() + rect.get_width(), rect.get_y()),
                        xytext=(15, -5),
                        textcoords="offset points",
                        )
    autolabel(rects1)

    axes = plt.gca()
    axes.set_xlim([0, times.max() * 1.2])
    # plt.barh(y_pos, pystencils, align="center", alpha=0.5)
    plt.yticks(index, functions)
    plt.xlabel("Execution time in seconds")
    plt.title("")
    plt.legend(loc='lower right')
    plt.tight_layout()
    plt.show()
    return avg_dict


def compare_q9_v17_v37(iterations, sizes):
    """Chpater Performance Comparison,
    Hash: 8b62f76c9cfb46aafebd4f52a6b5eecbfa62c56f

    Bar Plot comparison of q9 vs v17 vs v37 performance.
    """

    lattice_q9 = Lattice.from_name("D2Q9")
    lattice_v17 = Lattice.from_name("D2V17")
    lattice_v37 = Lattice.from_name("D2V37")

    time_dict = dict()
    for size in sizes:
        lbm9 = LBM(size=(size, size), lattice=lattice_q9, pystencils=True, omega=0.8, fill_mode="taylor-green")
        lbm17 = LBM(size=(size, size), lattice=lattice_v17, pystencils=True, omega=0.8, fill_mode="taylor-green")
        lbm37 = LBM(size=(size, size), lattice=lattice_v37, pystencils=True, omega=0.8, fill_mode="taylor-green")
        t1, t2, t3 = time_compare(lbm9.time_step, lbm17.time_step, lbm37.time_step, iters=iterations)
        time_dict[str(size)] = [t1/t1, t2/t1, t3/t1]
        print(time_dict[str(size)])

    # ======== Plot as bar chart
    import matplotlib.pyplot as plt;
    plt.rcdefaults()
    import numpy as np
    import matplotlib.pyplot as plt

    descriptions = ["Size ({} X {})".format(size, size) for size in sizes]
    times = np.zeros((3, len(descriptions)), dtype=np.float64)
    for i, size in enumerate(sizes):
        times[0, i] = time_dict[str(size)][0]
        times[1, i] = time_dict[str(size)][1]
        times[2, i] = time_dict[str(size)][2]

    q9_times = times[0, :]
    v17_times = times[1, :]
    v37_times = times[2, :]

    fig, ax = plt.subplots()
    index = np.arange(len(q9_times))
    bar_width = 0.2
    opacity = 0.8

    rects3 = plt.barh(index + bar_width, v37_times, bar_width- 0.02, alpha=opacity, color="purple", label="V37")
    rects2 = plt.barh(index, v17_times, bar_width - 0.02, alpha=opacity, color="blue", label="V17")
    rects1 = plt.barh(index - bar_width, q9_times, bar_width- 0.02, alpha=opacity, color="cyan", label="Q9")

    axes = plt.gca()
    axes.set_xlim([0, times.max() * 1.2])
    # plt.barh(y_pos, pystencils, align="center", alpha=0.5)
    plt.yticks(index, descriptions)
    plt.xlabel("Relative computation time")
    plt.title("")
    plt.legend()
    plt.tight_layout()
    plt.show()


def compare_np_ps_one_kernel():
    """Get Time per iteration for different sizes and lattices of the NumPy algorithm.

    Hash: 6a33b425359710cd75437c4524c9d132adb206f7


    Result:
    [[0.004118585586547851, 0.021164894104003906, 0.08900172710418701, 0.3898637771606445], [0.061525917053222655, 0.24237921237945556, 0.984809398651123, 3.9992587566375732]]
    """
    sizes = [128, 256, 512, 1024]
    iterations = 10
    lattices = ["D2Q9", "D2V17"]
    times = []
    for lat_name in lattices:
        lattice_times = []
        lattice = Lattice().from_name(lat_name)
        for size in sizes:
            print("{} Starting size: {}".format(lat_name, size))
            lbm = LBM((size, size), lattice, pystencils=False, omega=1.8181, hurry=True, fill_mode="taylor-green")
            time_start = time.time()
            for j in range(iterations):
                lbm.iterate()
            time_stop = time.time()
            lattice_times.append((time_stop - time_start)/iterations)
            if size == sizes[0] and lat_name == lattices[0]:
                print("Estimated overall time: {}".format((time_stop - time_start)*6*sum([(x/sizes[0])**2 for x in sizes])))

        times.append(lattice_times)
    print(times)


def compare_accuracy_ps_one_kernel():
    """
    Params for Velocity error:
    viscocity = 1/6
    u_0 = 0.04
    sizes = [128]
    error_type="ux"
    l2="root"
    iterations = 500

    """
    def compute_average(ar):
        """"""
        averages = np.zeros_like(ar, dtype=np.float64)
        averages[0] = ar[0]
        for i in np.arange(1, len(averages)):
            averages[i] = (i * averages[i-1] + ar[i])/(i+1)
        return averages

    viscocity = 1 / 200
    u_0 = 0.04
    sizes = [128]
    error_type = "density"
    l2 = "root"
    iterations = 1000

    import pystencils as ps
    lattices = ["D2Q9", "D2V17", "D2V37"]
    times = []
    lbms = []
    for lattice_name in lattices:
        time_by_lattice = []
        for size in sizes:
            print("{} Starting size: {}".format(lattice_name, size))
            lbm = LbmPs(lattice_name, viscocity=viscocity, resolution=size, iterations=iterations, u_0=u_0, calc_error=True, l2=l2, error_type=error_type)
            t = lbm.run_lbm()
            lbms.append(lbm)
            time_by_lattice.append(t)
        times.append(time_by_lattice)

    er9 = np.array(lbms[0].errors)
    er17 = np.array(lbms[1].errors)
    er37 = np.array(lbms[2].errors)

    print("Maximum error for Q9: {}".format(er9.max()))
    print("Maximum error for V17: {}".format(er17.max()))
    print("Maximum error for V37: {}".format(er37.max()))

    #filter
    er9 = [1 if e > 1 else e for e in er9]
    er17 = [1 if e > 1 else e for e in er17]
    er37 = [1 if e > 1 else e for e in er37]

    av9 = compute_average(er9)
    #av9 = np.ones_like(av9)
    #er9 = np.ones_like(er9)
    av17 = compute_average(er17)
    av37 = compute_average(er37)
    plt.gcf().subplots_adjust(left=0.3)

    plt.plot(lbms[0].times, er9, color="purple", alpha=0.3, linestyle='dashed')
    plt.plot(lbms[0].times, av9, color="purple", label="D2Q9")

    plt.plot(lbms[1].times, er17, color="blue", alpha=0.3, linestyle='dashed')
    plt.plot(lbms[1].times, av17, color="blue", label="D2V17")

    plt.plot(lbms[2].times, er37, color="black", alpha=0.3, linestyle='dashed')
    plt.plot(lbms[2].times, av37, color="black", label="D2V37")
    plt.ylabel(r"L2 error of density $\epsilon_{\rho}$")
    plt.xlabel("t")
    plt.legend()
    plt.show()


def stability(lattice_name, u_0_vals, viscosity_vals, iterations):
    """

    :param lattice_name:
    :param u_0_vals:
    :param relaxation_vals:
    :return:
    """

    import pystencils as ps
    size = 128
    lbm = LbmPs(lattice_name, viscocity=viscosity_vals[0], resolution=size, iterations=iterations, u_0=u_0_vals[0],
                calc_error=False, l2="root", error_type="ux", check_stability=True, silent=True)

    errors = np.ones((len(u_0_vals), len(viscosity_vals)), dtype=np.float64) * -1
    reynolds = np.zeros((len(u_0_vals), len(viscosity_vals)), dtype=np.float64)
    points = np.zeros((errors.shape) + (2,))
    for i, u_0 in enumerate(u_0_vals):
        for j, viscosity in enumerate(viscosity_vals):
            points[i, j] = [u_0, viscosity]
            if errors[i, j] != -1:
                continue
            lbm.cleanup(u_0, viscosity)
            reynolds[i, j] = lbm.reynold
            try:
                t = lbm.run_lbm()
            except StabilityException:
                errors[i, j:] = np.ones_like(errors[i, j])  # u_0 and higher is unstable
                #errors[i:, j] = np.ones_like(errors[i, j])  # u_0 and higher is unstable
                continue
            error = lbm.get_error(lbm.dh.gather_array("src"), iterations)
            errors[i, j] = min(error, 1)

    return errors



def stability_heatmap():
    iterations = 1000
    u_0_vals = np.arange(0.7, 0.25, -0.025)
    viscosity_vals = [Fraction(1, 6),
                      Fraction(1, 20),
                      Fraction(1, 60),
                      Fraction(1, 200),
                      Fraction(1, 600),
                      Fraction(1, 2000),
                      ]
    start_time = time.time()
    print("Starting Q9")
    errorsQ9 = stability(lattice_name="D2Q9",
                    u_0_vals=u_0_vals,
                    viscosity_vals=viscosity_vals,
                    iterations=iterations)
    print("Estimated total time {}s".format(9 * (time.time() - start_time)))
    est = time.time()

    print("Starting V17")
    errorsV17 = stability(lattice_name="D2V17",
                         u_0_vals=u_0_vals,
                         viscosity_vals=viscosity_vals,
                         iterations=iterations)

    print("Starting V37")
    errorsV37 = stability(lattice_name="D2V37",
                         u_0_vals=u_0_vals,
                         viscosity_vals=viscosity_vals,
                         iterations=iterations)

    print("Actual time: {}s".format(time.time() - est))
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    ax1 = plt.subplot(1, 3, 1)
    im1 = ax1.imshow(errorsQ9, cmap=plt.get_cmap("Blues_r"))

    ax2 = plt.subplot(1, 3, 2)
    im2 = ax2.imshow(errorsV17, cmap=plt.get_cmap("Blues_r"))

    ax3 = plt.subplot(1, 3, 3)
    im3 = ax3.imshow(errorsV37, cmap=plt.get_cmap("Blues_r"))

    ax1.set_xticks(np.arange(len(viscosity_vals)))
    ax1.set_yticks(np.arange(len(u_0_vals)))
    ax1.set_xticklabels([r"$\frac{" + str(v.numerator) + "}{" + str(v.denominator) + "}$" for v in viscosity_vals])
    ax1.set_yticklabels([round(u, 2) for u in u_0_vals])

    ax2.set_xticks(np.arange(len(viscosity_vals)))
    ax2.set_yticks(np.arange(len(u_0_vals)))
    ax2.set_xticklabels([r"$\frac{" + str(v.numerator) + "}{" + str(v.denominator) + "}$" for v in viscosity_vals])
    ax2.set_yticklabels([round(u, 2) for u in u_0_vals])

    ax3.set_xticks(np.arange(len(viscosity_vals)))
    ax3.set_yticks(np.arange(len(u_0_vals)))
    ax3.set_xticklabels([r"$\frac{" + str(v.numerator) + "}{" + str(v.denominator) + "}$" for v in viscosity_vals])
    ax3.set_yticklabels([round(u, 2) for u in u_0_vals])

    ax1.set_ylabel(r"Maximum velocity $u_0$")
    ax2.set_xlabel(r"Viscosity $\nu$")

    ax1.set_title("D2Q9")
    ax2.set_title("D2V17")
    ax3.set_title("D2V37")
    #divider = make_axes_locatable(ax3)
    #cax = divider.append_axes("right", size="5%", pad=0.05)

    #plt.colorbar(im3, fraction=0.040, pad=0.04)
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    # logger.setLevel(logging.INFO)
    logger.setLevel(logging.WARNING)

    # Visualizations:

    # compare_numpy_accuracy()
    # compare_v37_pystencils()
    # compare_v37_accuracy()
    # performance_numpy_vs_pysten
    #     # compare_numpy_performance()cils_v17(average_over=[64, 128, 256, 512], iterations=100)
    # compare_accuracy_ps_one_kernel()
    # compare_np_ps_one_kernel()

    # stability_heatmap()
    compare_accuracy_ps_one_kernel()

    # compare_q9_v17_v37(100, sizes=[64, 128, 256, 512])
