from lbmweights import Lattice
from pystencils.session import *



def run_lbm(lattice_name, tau, resolution, iterations):
    def l2norm(approximation, solution):
        """
        Calculating the l2norm of two matrices.
        The matrices have to have the same shape

        :return Value that indicates how close mat2 i
        """
        assert (
                approximation.shape == solution.shape
        ), "The shape of the matrices have to be the same"

        diff = ((approximation - solution) ** 2).sum()
        abs = (solution ** 2).sum()
        norm = diff / abs
        return norm

    def taylor_green(data_shape, t, tau, c_s_sq, test=False):
        N_x = data_shape[0]
        N_y = data_shape[1]
        u_0 = 0.03
        p_0 = c_s_sq
        roh_avg = 0.1

        # Reynolds configs
        viscocity = c_s_sq * (tau - 0.5)

        kx = 2 * np.pi
        ky = 2 * np.pi

        assert abs(c_s_sq * (tau - 0.5) - viscocity) <= 1e-10

        X, Y = np.meshgrid(
            np.linspace(0, 1, num=N_x, endpoint=True),
            np.linspace(0, 1, num=N_y, endpoint=True),
            indexing='ij'
        )

        t_d = 1 / (viscocity * ((2 * np.pi / N_x) ** 2 + (2 * np.pi / N_y) ** 2))
        u_x = u_0 * (-1 * np.cos(kx * X) * np.sin(ky * Y)) * np.exp(- t / t_d)
        u_y = u_0 * (np.sin(kx * X) * np.cos(ky * Y)) * np.exp(- t / t_d)
        rho = (p_0 - roh_avg * u_0 ** 2 / 4 * (np.cos(2 * kx * X) + np.cos(2 * ky * Y)) * np.exp(-2 * t / t_d)) / c_s_sq

        return rho, u_x, u_y

    def macroscopic_moments_kernel(lattice):
        c_s_sq, weights, velocities = lattice.velocity_set()
        Q = lattice.q
        velocities = lattice.velocities

        c_ix = [vel[0] for vel in velocities]
        c_iy = [vel[1] for vel in velocities]

        src = ps.fields("src({Q}): double[2D]".format(Q=Q))
        velocity = ps.fields("velocity(2): [2D]")
        density, = sp.symbols("density, ")

        density_formula = sum([src[0, 0](i) for i in range(Q)])
        vel_x_formula = (sum([src[0, 0](i) * c_ix[i] for i in range(Q)])) / density
        vel_y_formula = (sum([src[0, 0](i) * c_iy[i] for i in range(Q)])) / density

        symbolic_description = [ps.Assignment(density, density_formula),
                                ps.Assignment(velocity[0, 0](0), vel_x_formula),
                                ps.Assignment(velocity[0, 0](1), vel_y_formula),
                                ]
        return ps.create_kernel(symbolic_description).compile()

    def init_kernel(lattice):
        c_s_sq, weights, velocities = lattice.velocity_set()
        Q = lattice.q
        velocities = lattice.velocities
        weights = [complex(x).real for x in weights]
        c_ix = [vel[0] for vel in velocities]
        c_iy = [vel[1] for vel in velocities]

        density = ps.fields("density: double[2D]")
        u_x = ps.fields("u_x: double[2D]")
        u_y = ps.fields("u_y: double[2D]")
        dst = ps.fields("dst({Q}): double[2D]".format(Q=Q))

        sym = []
        for i in range(Q):
            feq_formula = weights[i] * density[0, 0] * (
                    1
                    + (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) / c_s_sq
                    + (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 2 / (2 * c_s_sq ** 2)
                    - (u_x[0, 0] ** 2 + u_y[0, 0] ** 2) / (2 * c_s_sq)
            )

            if lattice.eq_order >= 3:
                feq_formula += weights[i] * density[0, 0] * (
                        -1 / (2 * c_s_sq ** 2) * (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) * (
                            u_x[0, 0] ** 2 + u_y[0, 0] ** 2)
                        + 4 / 3 * (1 / (2 * c_s_sq)) ** 3 * (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 3
                )
            if lattice.eq_order >= 4:
                feq_formula += weights[i] * density[0, 0] * (1 / (6 * c_s_sq ** 2) * (
                        (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 4 / (4 * c_s_sq ** 2) - (
                        3 * (u_x[0, 0] ** 2 + u_y[0, 0] ** 2) *
                        (u_x[0, 0] * c_ix[i] + u_y[0, 0] * c_iy[i]) ** 2) / (2 * c_s_sq) + 3 * (
                                u_x[0, 0] ** 4 + u_y[0, 0] ** 4)))

            sym.append(ps.Assignment(dst[0, 0](i), feq_formula))
        return ps.create_kernel(sym).compile()

    def get_error(data, time, lattice):
        rho, u_x, u_y = taylor_green(size, time, tau, float(lattice.c_s_sq))
        velocities = np.zeros(size + (2,), dtype=np.float64)
        upd_mac(src=np.copy(data), velocity=velocities)
        abs_vel = velocities[:, :, 0] ** 2 + velocities[:, :, 1] ** 2
        abs_ana = u_x ** 2 + u_y ** 2
        return l2norm(abs_vel, abs_ana)

    def vel_length(field):
        velocities = np.zeros(size + (2,), dtype=np.float64)
        upd_mac(src=np.copy(field), velocity=velocities)
        abs_vel = velocities[:, :, 0] ** 2 + velocities[:, :, 1] ** 2
        return abs_vel

    def initialize(dh):
        init_ker = init_kernel(lattice)
        rho, u_x, u_y = taylor_green(size, 0, tau, float(lattice.c_s_sq))

        entries_per_cell = dh.gather_array("src").shape[-1]
        init_feq = np.zeros((1,) + size + (entries_per_cell,))
        init_ker(density=rho, u_x=u_x, u_y=u_y, dst=init_feq[0])

        for block in dh.iterate(ghost_layers=False):
            x, y = block.cell_index_arrays
            block["src"][True] = init_feq
            pdfAC = dh.gather_array("src", ghost_layers=False)


    lattice = Lattice.from_name(lattice_name)
    size = (resolution, resolution)
    assignment_collection = lattice.assignment_collection

    upd_mac = macroscopic_moments_kernel(lattice)
    dh = ps.create_data_handling(size, periodicity=(True, True))
    src = dh.add_array('src', values_per_cell=lattice.q, ghost_layers=lattice.necessary_ghost_layers)
    dst = dh.add_array_like('dst', 'src')
    copy_func = dh.synchronization_function(['src'], stencil='D2Q9')
    kernel = ps.create_kernel(assignment_collection).compile()
    initialize(dh)

    it = iterations
    times = [0,]
    errors = [0,]
    import time
    timing = time.time()
    for i in np.arange(1, it + 1):
        copy_func()
        dh.run_kernel(kernel, omega=1/tau)
        dh.swap('src', 'dst')
        #errors.append(get_error(dh.gather_array("src", ghost_layers=False), time=i, lattice=lattice))
        #times.append(i)

    total_time = time.time() - timing
    return total_time / iterations

#plt.plot(times, errors)
#plt.show()


sizes = [128, 256, 512, 1028]
lattices = ["D2Q9", "D2V17"]
times = []
for lattice_name in lattices:
    time_by_lattice = []
    for size in sizes:
        print("{} Starting size: {}".format(lattice_name, size))
        t = run_lbm(lattice_name, 0.55, size, 10)
        time_by_lattice.append(t)
    times.append(time_by_lattice)
print(times)
